part of masamune_notion;

final notionPageCollectionProvider =
    ChangeNotifierProvider.family<NotionPageCollectionModel, String>(
  (_, databaseId) => NotionPageCollectionModel(databaseId),
);

class NotionPageCollectionModel
    extends ValueModel<List<NotionPageDocumentModel>>
    with ListModelMixin<NotionPageDocumentModel>
    implements DynamicCollectionModel<NotionPageDocumentModel> {
  NotionPageCollectionModel(this.databaseQuery) : super([]);

  final String databaseQuery;
  // ignore: unused_field
  String? _nextCursor;
  FirestoreDynamicDocumentModel? _document;
  FirestoreDynamicCollectionModel? _collection;

  /// It becomes `true` after [loadOnce] is executed.
  bool loaded = false;

  @override
  Future<void>? get loading => _loadCompleter?.future;
  Completer<void>? _loadCompleter;

  @override
  Future<void>? get saving => throw UnimplementedError();

  /// Get an instance of Firebase Functions.
  @protected
  FirebaseFunctions get functions {
    return FirebaseFunctions.instanceFor(region: FirebaseCore.region);
  }

  /// If this value is true,
  /// Notify changes when there are changes in the list itself using list-specific methods.
  @override
  bool get notifyOnChangeList => false;

  /// If this value is true,
  /// the change will be notified when [value] itself is changed.
  @override
  bool get notifyOnChangeValue => true;

  /// Callback before the load has been done.
  @protected
  @mustCallSuper
  Future<void> onLoad() async {}

  /// Callback after the load has been done.
  @protected
  @mustCallSuper
  Future<void> onDidLoad() async {}

  /// You can describe the process when you get the [response].
  ///
  /// Describe error handling (throw Exception in case of error), etc.
  @protected
  @mustCallSuper
  @protected
  @mustCallSuper
  void onCatchResponse(HttpsCallableResult<Map> response) {}

  /// Add a process to create a document object.
  @protected
  NotionPageDocumentModel createDocument() {
    throw UnimplementedError();
  }

  /// Create a new document.
  NotionPageDocumentModel create() {
    throw UnimplementedError();
  }

  /// Functions can be called to retrieve the data.
  ///
  /// You can specify the Post data to be called in [parameters].
  Future<NotionPageCollectionModel> load([bool useCache = true]) async {
    if (_loadCompleter != null) {
      await loading;
      return this;
    }
    if (!NotionCore.initialized) {
      throw Exception(
        "Notion settings have not been initialized. Please initialize it by executing [NotionCore.initialize()].",
      );
    }
    _loadCompleter = Completer();
    try {
      if (value.isNotEmpty) {
        _loadCompleter?.complete();
        _loadCompleter = null;
        return this;
      }
      final _query = NotionQuery.fromPath(databaseQuery);
      final databaseId =
          _query.path.substring(_query.path.length - 32, _query.path.length);
      final queryKey = databaseQuery.toBase64();
      final query = NotionCore._defaultQuery != null
          ? NotionQueryGroup([
              _query,
              NotionCore._defaultQuery!,
            ])
          : _query;
      await FirebaseCore.initialize();
      await onLoad();
      _document ??= readProvider(
        firestoreDocumentProvider("${NotionCore._indexPath}/$queryKey"),
      );
      await _document?.loadOnce();
      final indexList = _document.getAsList<String>("index").distinct();
      _collection ??= readProvider(
        firestoreCollectionProvider(
          ModelQuery(
            NotionCore._contentPath,
            key: "uid",
            whereIn: indexList,
          ).value,
        ),
      );
      await _collection?.loadOnce();
      clear();
      final data = indexList.mapAndRemoveEmpty((uid) {
        final e = _collection.firstWhereOrNull((item) => item.uid == uid);
        if (e == null) {
          return null;
        }
        final doc = NotionPageDocumentModel(uid, {});
        doc.value = doc.fromMap(e);
        return doc;
      });
      addAll(data);
      await onDidLoad();
      if (isNotEmpty) {
        notifyListeners();
        _loadCompleter?.complete();
        _loadCompleter = null;
      }
      final filter = query.toFilter();
      final res =
          await functions.httpsCallable(NotionCore._endpoint).call<Map>({
        "type": "database",
        "id": databaseId,
        "key": queryKey,
        if (filter != null) "filter": filter,
      });
      final map = res.data.cast<String, dynamic>();
      final list = map.getAsList("results").cast<DynamicMap>();
      if (isEmpty && list.isNotEmpty) {
        final data = list.mapAndRemoveEmpty((page) {
          final uid = page.get("id", "").replaceAll("-", "");
          if (uid.isEmpty) {
            return null;
          }
          final doc = NotionPageDocumentModel(uid, {});
          doc.value = doc.fromMap(page);
          return doc;
        });
        addAll(data);
        notifyListeners();
      }
      _loadCompleter?.complete();
      _loadCompleter = null;
    } catch (e) {
      _loadCompleter?.completeError(e);
      _loadCompleter = null;
      rethrow;
    } finally {
      _loadCompleter?.complete();
      _loadCompleter = null;
    }
    return this;
  }

  /// Reload data and updates the data in the model.
  ///
  /// It is basically the same as the [load] method,
  /// but combining it with [loadOnce] makes it easier to manage the data.
  Future<NotionPageCollectionModel> reload([bool useCache = true]) =>
      load(useCache);

  /// If the data is empty, [load] is performed only once.
  ///
  /// In other cases, the value is returned as is.
  ///
  /// Use [isEmpty] to determine whether the file is empty or not.
  Future<NotionPageCollectionModel> loadOnce([bool useCache = true]) async {
    if (!loaded) {
      loaded = true;
      return load(useCache);
    }
    return this;
  }
}
